from time import time, sleep
from leer.wallet.key_manager import KeyManagerClass

from leer.core.lubbadubdub.ioput import IOput
from leer.core.lubbadubdub.address import Address
from leer.core.lubbadubdub.transaction import Transaction

from uuid import uuid4
import logging
import base64
     


logger = logging.getLogger("Wallet")

def wallet(syncer, config):
  '''
    Wallet is synchronous service which holds private keys and information about 
    owned outputs. It provides information for transactions and block templates
    generation.
  '''
  def get_height():
    _id = str(uuid4())
    syncer.queues['Notifications'].put({'action':'get', 'id':_id, 'key': 'blockchain height','sender': "Wallet"})
    message_queue = syncer.queues['Wallet']
    while True:
      put_back = [] #We wait for specific message, all others will wait for being processed
      while not message_queue.empty():
        message = message_queue.get()
        if (not 'id' in message)  or (not message['id']==_id):
          put_back.append(message)
        if message['result']=='error':
          raise KeyError
        return message['result']['value']
      sleep(0.01)
      for message in put_back:
        message_queue.put(message)

  message_queue = syncer.queues['Wallet']
  _path = config['location']['wallet']
  km = KeyManagerClass(path=_path)
  while True:
    sleep(0.01)
    while not message_queue.empty():
      message = message_queue.get()
      logger.info("Process message %s"% message)
      if not 'action' in message:
        continue
      if message['action']=="process new block":
        tx = Transaction(txos_storage=None)
        tx.deserialize(message['tx'],skip_verification=True)
        block_height = message['height']
        for index in tx.inputs:
          if km.is_unspent(index): #Note it is not check whether output is unspent or not, we check that output is marked as our and unspent in our wallet
            km.spend_output(index, block_height)
        for _o in tx.outputs:
          if km.is_owned_pubkey(_o.address.pubkey.serialize()):
            km.add_output(_o, block_height)
      if message['action']=="process rollback":
        rollback = message['rollback_object']
        block_height = message['block_height']
        km.rollback(block_height)
      if message['action']=="process indexed outputs": #during private key import correspondent outputs will be processed again
        pass
      if message['action']=="give new taddress":
        address = km.new_address()
        response = {"id": message["id"], "result": address.to_text()}
        syncer.queues[message['sender']].put(response)
      if message['action']=="give new address":
        address = km.new_address()
        response = {"id": message["id"], "result": address.serialize()}
        syncer.queues[message['sender']].put(response)
      if message['action']=="get confirmed balance stats":
        response = {"id": message["id"]}
        try:
          height = get_height()
          stats = km.get_confirmed_balance_stats(height)
          response["result"] = stats
        except KeyError:
          response["result"] = "error: core_loop didn't set height yet"
        syncer.queues[message['sender']].put(response)
      if message['action']=="get confirmed balance list":
        response = {"id": message["id"]}
        try:
          height = get_height()
          stats = km.get_confirmed_balance_list(height)
          response["result"] = stats
        except KeyError:
          response["result"] = "error: core_loop didn't set height yet"
        syncer.queues[message['sender']].put(response)
      if message['action']=="give private key":
        pass
      if message['action']=="take private key":
        pass
      if message['action']=="generate tx template":
        response = {"id": message["id"]}
        value  = int(message["value"])
        taddress = message["address"]
        a = Address()
        a.from_text(taddress)
        try:
          current_height = get_height()
        except KeyError:
           response["result"] = "error"
           response["error"] = "core_loop didn't set height yet"
           syncer.queues[message['sender']].put(response)
           continue
        _list = km.get_confirmed_balance_list(current_height)
        list_to_spend = []
        summ = 0 
        utxos = []
        for address in _list:
            for texted_index in _list[address]:
              if summ>value+100000000: #TODO fee here
                continue
              if isinstance(_list[address][texted_index], int):
                _index = base64.b64decode(texted_index.encode())
                utxos.append(_index)
                summ+=_list[address][texted_index]
        if summ < value:
            response["result"] = "error"
            response["error"] = "Not enough matured coins"
            syncer.queues[message['sender']].put(response)
            continue
        
        tx_template = { 'priv_by_pub': {}, 'change address': km.new_address().serialize(), 'utxos':utxos,
                        'address': a.serialize(), 'value': value }
        for utxo in utxos:
          pub,priv = km.priv_and_pub_by_output_index(utxo)
          tx_template['priv_by_pub'][pub]=priv
        
        response["result"]=tx_template
        syncer.queues[message['sender']].put(response)
      if message['action']=="stop":
        return

    




from secp256k1_zkp import PublicKey, PrivateKey
from leer.core.lubbadubdub.constants import GLOBAL_TEST

from leer.core.lubbadubdub.address import Address

'''
#signleton
class KeyManagerClass:
  def __init__(self):
    if not GLOBAL_TEST['key manager']:
      raise NotImplemented
    self.memory={}

  def new_address(self):
    r=PrivateKey()
    self.memory[r.pubkey.serialize()]=r
    return Address().from_private_key(r)

  def priv_by_pub(self, pubkey):
    return self.memory[pubkey.serialize()]

  def priv_by_address(self, address):
    try:
      return self.memory[address.pubkey.serialize()]
    except KeyError:
      return False

  def add_privkey(self, privkey):
    self.memory[privkey.pubkey.serialize()]=privkey

KeyManager=KeyManagerClass()
'''
'''
#signer
KeyManager.add_privkey(PrivateKey(b'%k3\xbe\xb4F\xb1"\x0e\xdd\xad:bC\xae\x98\xbaX\x96dgl\x1f~\xbb\xcdH\x98\x83\xcd&\x9f', raw=True))

#genesis
KeyManager.add_privkey(PrivateKey(b'\xbe\xf2%\x18\xefU\x11\xc3\x86\xe0?.\x8d\xd3\xdf\xb8\xae+\xc2|\x98\x82\xf50\x89>.\xa6\x07\x10\x841', raw=True))
KeyManager.add_privkey(PrivateKey(b'\xbe\xf2%\x18\xefU\x12\xc3\x86\xe0?.\x8d\xd3\xdf\xb8\xae+\xc2|\x98\x82\xf50\x89>.\xa6\x07\x10\x841', raw=True))
KeyManager.add_privkey(PrivateKey(b'\xbe\xf2%\x18\xefU\x13\xc3\x86\xe0?.\x8d\xd3\xdf\xb8\xae+\xc2|\x98\x82\xf50\x89>.\xa6\x07\x10\x841', raw=True))
KeyManager.add_privkey(PrivateKey(b'\xbe\xf2%\x18\xefU\x14\xc3\x86\xe0?.\x8d\xd3\xdf\xb8\xae+\xc2|\x98\x82\xf50\x89>.\xa6\x07\x10\x841', raw=True))

'''



from secp256k1_zkp import PrivateKey
from leer.core.storage.default_paths import key_manager_path
import shutil, os, time, lmdb, math
from hashlib import sha256
from leer.core.lubbadubdub.address import address_from_private_key
import base64

class KeyManagerClass:
  def __init__(self, password=None, path = key_manager_path):
    #self.privkey = None
    #if password:
    #  password_bytes = bytes(password+")dod8q=(3fnS#904ff", "utf-16")
    #  m = sha256()
    #  m.update(password_bytes)
    #  self.privkey = Privkey(m.digest(), raw=True)
    self.wallet = DiscWallet(dir_path=path)

  def new_address(self):
    prk=PrivateKey()
    self.wallet.add_privkey_to_pool(prk.pubkey.serialize(), prk.private_key)
    privkey = PrivateKey(self.wallet.get_privkey_from_pool(), raw=True)
    return address_from_private_key(privkey) 

  def priv_by_pub(self, pubkey):
    raw_priv = self.wallet.get_privkey(pubkey.serialize())
    if not raw_priv:
      raise KeyError("Private key not in the wallet")
    return PrivateKey(raw_priv, raw=True)

  def priv_by_address(self, address):
    raw_priv = self.wallet.get_privkey(address.pubkey.serialize())
    if not raw_priv:
      raise KeyError("Private key not in the wallet")
    return PrivateKey(raw_priv, raw=True)

  def add_privkey(self, privkey):
    self.wallet.put(privkey.pubkey.serialize(), privkey.private_key)

  def fill_pool(self, keys_number=100):
    for _ in range(keys_number):
     prk=PrivateKey()
     self.wallet.add_privkey_to_pool(prk.pubkey.serialize(), prk.private_key)

  @property
  def pool_size(self):
    raw_pool_size = self.wallet.get_pool_size()
    if raw_pool_size:
      return int.from_bytes(self.wallet.get_pool_size(), "big") 
    else: #NoneType
      return 0

  def get_confirmed_balance_stats(self, utxo_index, txos_storage, current_height):
    stats = {
              'matured': {'known_value':0, 'known_count':0, 'unknown_count':0},
              'immatured': {'known_value':0, 'known_count':0, 'unknown_count':0}
            }
    with self.wallet.env.begin(write=False) as txn:
      cursor = txn.cursor(db=self.wallet.main_db)
      for ser_pub in cursor.iternext(keys=True, values=False):
        for output_index in utxo_index.get_all_unspent_for_serialized_pubkey(ser_pub):
          try:
             lock_height, value, serialized_index = self.wallet.get_output(output_index, txn=txn)
          except KeyError:
            utxo = txos_storage.confirmed[output_index]
            lock_height = utxo.lock_height
            serialized_index = utxo.serialized_index
            if utxo.detect_value(self):
               value = utxo.value
            else:
               value = None
            self.wallet.put_output(output_index, (lock_height, value, serialized_index))#No txn here, write-access is required
          if value:
            if current_height>=lock_height:
              stats['matured']['known_value']+=value
              stats['matured']['known_count']+=1
            else:
              stats['immatured']['known_value']+=value
              stats['immatured']['known_count']+=1
          else:
            if current_height>=lock_height:
              stats['matured']['unknown_count']+=1
            else:
              stats['immatured']['unknown_count']+=1
    return stats

  def get_confirmed_balance_list(self, utxo_index, txos_storage, current_height):
    ret = {}
    with self.wallet.env.begin(write=False) as txn:
      cursor = txn.cursor(db=self.wallet.main_db)
      for ser_pub, priv in cursor.iternext(keys=True, values=True):
        address = address_from_private_key(PrivateKey(priv, raw=True))
        taddress = address.to_text()
        for output_index in utxo_index.get_all_unspent_for_serialized_pubkey(ser_pub):
          try:
             lock_height, value, serialized_index = self.wallet.get_output(output_index, txn=txn)
          except KeyError:
            utxo = txos_storage.confirmed[output_index]
            lock_height = utxo.lock_height
            serialized_index = utxo.serialized_index
            if utxo.detect_value(self):
               value = utxo.value
            else:
               value = None
            self.wallet.put_output(output_index, (lock_height, value, serialized_index))#No txn here, write-access is required
          texted_index = base64.b64encode(serialized_index).decode()
          if not taddress in ret:
            ret[taddress]={}
          if value:
            ret[taddress][texted_index]=value
          else:
            ret[taddress][texted_index]='unknown'
    return ret

def serialize_output_params(p):
  lock_height, value, serialized_index = p
  ser_lock_height = lock_height.to_bytes(4,"big")
  if value == None:
    ser_value = b"\xff"*7
  else:
    ser_value = value.to_bytes(7,"big")
  return ser_lock_height+ser_value+serialized_index

def deserialize_output_params(p):
  lock_height = int.from_bytes(p[:4], "big")
  value = int.from_bytes(p[4:11], "big")
  serialized_index = p[11:]
  if value == 72057594037927935: #=b"\xff"*7
    value = None
  return lock_height, value, serialized_index


class DiscWallet:
  '''
    It is generally key-value db with two types of records:
     1) key is serialized pubkey, value - serialized privkey.
     2) key is output_index, value - tuple (lock_height, value)
    There is privkey pool: bunch of pregenerated privkeys. It is expected that on a higher level
    instead of generating and immediate usage of new key, new key will be put into the pool and the oldest key
    from the pool will be used. Thus, in case of backups, copies and so on, "old copy" will contain
    some keys used after copy being made.
  '''
  def __init__(self, dir_path):
    self.dir_path = dir_path

    if not os.path.exists(self.dir_path): 
        os.makedirs(self.dir_path) #TODO catch
    self.env = lmdb.open(self.dir_path, max_dbs=10, map_size=int(250e6)) #TODO this database is too big, should be checked
    with self.env.begin(write=True) as txn:
      self.main_db = self.env.open_db(b'main_db', txn=txn, dupsort=False)
      self.pool = self.env.open_db(b'pool', txn=txn, dupsort=False)
      self.output = self.env.open_db(b'output', txn=txn, dupsort=False)
      #if not txn.get(b'pool_size', db=self.pool):
      #  txn.put( b'pool_size', 0, db=self.pool) 



  def get_pool_prop(self, prop, txn = None):
    if not txn:
      with self.env.begin(write=False) as txn:
        return self.get_pool_prop(prop, txn)
    else:
      return txn.get( prop, db=self.pool) 

  def get_pool_size(self,txn=None):
    return self.get_pool_prop(b'size', txn=txn)

  def get_pool_current(self,txn=None):
    return self.get_pool_prop(b'current', txn=txn)

  def get_pool_top(self,txn=None):
    return self.get_pool_prop(b'top', txn=txn)

  def set_pool_prop(self, prop, value, txn = None):
    if not txn:
      with self.env.begin(write=True) as txn:
        return self.set_pool_prop(prop, value)
    else:
      if not value==None:
        return txn.put( prop, value.to_bytes(4,'big'), db=self.pool) 
      else:
        return txn.delete( prop, db=self.pool) 

  def set_pool_size(self, value, txn=None):
    return self.set_pool_prop(b'size', value, txn=txn)

  def set_pool_current(self, value, txn=None):
    return self.set_pool_prop(b'current', value, txn=txn)

  def set_pool_top(self, value, txn=None):
    return self.set_pool_prop(b'top', value, txn=txn)



  def put(self, serialized_pubkey, serialized_privkey, txn=None):
    if not txn:
      with self.env.begin(write=True) as txn:
        self.put(serialized_pubkey, serialized_privkey, txn=txn)
    else:
      p1=txn.put( bytes(serialized_pubkey), bytes(serialized_privkey), db=self.main_db)    
  

  def put_output(self, output_index, output_params, txn=None):
    if not txn:
      with self.env.begin(write=True) as txn:
        self.put_output(output_index, output_params, txn=txn)
    else:     
      p1=txn.put( bytes(output_index), serialize_output_params(output_params), db=self.output)    

  def get_output(self, output_index, txn=None):
    if not txn:
      with self.env.begin(write=False) as txn:
        self.get_output(output_index, txn=txn)
    else:     
      output_params = txn.get( bytes(output_index), db=self.output)    
      if not output_params:
         raise KeyError
      else:
        return deserialize_output_params( output_params)

  def get_privkey_from_pool(self):
    '''
      Privkey will be immideately removed from the pool.
    '''
    with self.env.begin(write=True) as txn:
        # pool is integer->pubkey key_value
        current = self.get_pool_current(txn=txn)
        if not current:
          raise Exception("Pool is empty")
        privkey = txn.get( txn.get(current, db=self.pool), db=self.main_db)
        current_int = int.from_bytes(current, 'big')
        top_int = int.from_bytes(self.get_pool_top(txn=txn), 'big')
        if current_int < top_int:
          self.set_pool_current(current_int+1, txn=txn)
        else:
          self.set_pool_current(None, txn=txn)
        self.set_pool_size(int.from_bytes(self.get_pool_size(txn=txn), 'big')-1, txn=txn)
        return privkey
        

  def add_privkey_to_pool(self, serialized_pubkey, serialized_privkey):
    with self.env.begin(write=True) as txn:
      self.put(serialized_pubkey, serialized_privkey, txn=txn)
      pool_current = self.get_pool_current(txn=txn)
      if pool_current == None:
         self.set_pool_size(1, txn=txn)
         pool_current = b"\x00\x00\x00\x00"
         current = 0
         self.set_pool_current(current, txn=txn)
         self.set_pool_top(current, txn=txn)
      else:
         self.set_pool_size(int.from_bytes(self.get_pool_size(txn=txn),"big")+1, txn=txn)
         self.set_pool_top(int.from_bytes(self.get_pool_top(txn=txn),"big")+1, txn=txn)
      txn.put( self.get_pool_top(txn=txn), serialized_pubkey, db=self.pool)  


  def get_privkey(self, serialized_pubkey):
    '''
      Note: Being get by pubkey, key pair is not checked wether it was in the pool or not.
    '''
    with self.env.begin(write=False) as txn:
      return txn.get(serialized_pubkey, db=self.main_db)
  

 



from leer.core.primitives.header import Header, ContextHeader
import shutil, os, time, lmdb, math
from leer.core.storage.default_paths import headers_storage_path

class HeadersStorage:

  __shared_states = {}

  def __init__(self, storage_space, path):
    if not path in self.__shared_states:
        self.__shared_states[path]={}
    self.__dict__ = self.__shared_states[path]
    self.storage = HeadersDiscStorage(path)
    self.storage_space = storage_space
    self.storage_space.register_headers_storage(self)

  def __getitem__(self, _hash):
    serialized_header = self.storage.get_by_hash(_hash)
    if not serialized_header:
      raise KeyError(_hash)
    ch=ContextHeader()
    ch.deserialize(serialized_header)
    return ch

  def __setitem__(self, _hash, header):
    #here we should save
    self.storage.put(header.height, header.hash, header.serialize_with_context())

  def update(self, _hash, header):
    self.storage.update(header.hash, header.serialize_with_context())

  def __contains__(self, _hash):
    return bool(self.storage.get_by_hash(_hash))
      
  def get_headers_at_height(self, height):
    ret=[]
    for serialized_header in self.storage.get_by_height(height):
      ch=ContextHeader()
      ch.deserialize(serialized_header)
      ret.append(ch)
    return ret

  def get_headers_hashes_at_height(self, height):
    return self.storage.get_hashes_by_height(height)


def __(x):
  return (x).to_bytes(4,'big')

class HeadersDiscStorage:
  def __init__(self, dir_path):
    self.dir_path = dir_path

    if not os.path.exists(self.dir_path): 
        os.makedirs(self.dir_path) #TODO catch
    self.env = lmdb.open(self.dir_path, max_dbs=10)
    with self.env.begin(write=True) as txn:
      self.main_db = self.env.open_db(b'main_db', txn=txn, dupsort=False)
      self.height_db = self.env.open_db(b'height_db', txn=txn, dupsort=True)

  def put(self, height, _hash, serialized_header):
    with self.env.begin(write=True) as txn:
      p1=txn.put( bytes(_hash), bytes(serialized_header), db=self.main_db, dupdata=False, overwrite=True)
      p2=txn.put( __(height), bytes(_hash), db=self.height_db, dupdata=True)

  def update(self, _hash, serialized_header):
    with self.env.begin(write=True) as txn:
      txn.put(bytes(_hash), bytes(serialized_header), db=self.main_db, dupdata=False, overwrite=True)

  def get_by_hash(self, _hash):
    with self.env.begin(write=False) as txn:
      return txn.get(bytes(_hash), db=self.main_db)

  def get_by_height(self, height):
    with self.env.begin(write=False) as txn:
      cursor = txn.cursor(db=self.height_db)
      assert cursor.set_key(__(height))
      _hashes = list(cursor.iternext_dup())
      #we do not use self.get_by_hash to keep one txn. Consider adding optional txn for all funcs?
      return [txn.get(bytes(_hash), db=self.main_db) for _hash in _hashes] 
  
  def get_hashes_by_height(self, height):
    with self.env.begin(write=False) as txn:
      cursor = txn.cursor(db=self.height_db)
      assert cursor.set_key(__(height))
      return list(cursor.iternext_dup())
  




from secp256k1_zkp import PrivateKey
from hashlib import sha256

def mining_canary_hash(_bytes):
    m = sha256()
    m.update(_bytes)
    p = PrivateKey(privkey=m.digest(), raw=True)
    m2 = sha256()
    m2.update(p.pubkey.serialize()[1:])
    return m2.digest()



from leer.core.storage.txos_storage import TXOsStorage
from leer.core.storage.default_paths import blocks_storage_path
from leer.core.primitives.block import Block, ContextBlock
import shutil, os, time, lmdb, math

class BlocksStorage:
  __shared_states = {}
  def __init__(self, storage_space, path):
    if not path in self.__shared_states:
      self.__shared_states[path]={}
    self.__dict__ = self.__shared_states[path]

    self.storage = BlocksDiscStorage(path)
    self.storage_space = storage_space
    self.storage_space.register_blocks_storage(self)
    self.download_queue = []
    
  def __getitem__(self, _hash):
    if not _hash in self.storage:
      raise KeyError(_hash)
    serialized_context_block = self.storage.get_by_hash(_hash)
    block=Block(storage_space=self.storage_space)
    cblock = ContextBlock(block=block)
    cblock.deserialize_raw(serialized_context_block)
    return cblock

  def __setitem__(self, _hash, block):
    #here we should save
    self.storage.put(_hash, block.serialize_with_context())

  def __contains__(self, _hash):
    return _hash in self.storage  

  def is_block_downloaded(self, _hash, auto_download=True):
    asked = False
    result = True
    block = self[_hash]
    for output in block.transaction_skeleton.output_indexes:
      if not self.storage_space.txos_storage.known(output):
        result = False
        if auto_download:
          self._ask_for_txout(output)
          asked = True
    if asked:
      self._ask_for_txouts()
    return result

  def get_rollback_object(self, _hash):
    serialized_rollback = self.storage.get_rollback_object(_hash)
    rb=RollBack()
    rb.deserialize_raw(serialized_rollback)
    return rb

  def pop_rollback_object(self, _hash):
    serialized_rollback = self.storage.pop_rollback_object(_hash)
    rb=RollBack()
    rb.deserialize_raw(serialized_rollback)
    return rb

  def put_rollback_object(self, _hash, rollback):
    self.storage.put_rollback_object(_hash, rollback.serialize())



  def _ask_for_txout(self, txout):
    if not txout in self.download_queue:
      self.download_queue.append(txout)

  def _ask_for_txouts(self):
    self.ask_for_txouts_hook(self.download_queue)
    self.download_queue = []

  def forget_block(self, _hash):
    self.storage.delete_block_by_hash(_hash)



class RollBack:
  def __init__(self):
    self.pruned_inputs = [] #revert objects
    self.num_of_added_outputs = None
    self.num_of_added_excesses = None
    self.prev_state = None

  def serialize_bytes(self, _bytes):
    return len(_bytes).to_bytes(2,"big")+_bytes

  def deserialize_bytes_raw(self, serialized):
    _slen, serialized = serialized[:2], serialized[2:]
    _len = int.from_bytes(_slen, "big")
    _bytes, serialized = serialized[:_len], serialized[_len:]
    return _bytes, serialized

  def serialize(self):
    # each pruned input is (TXOS[num, _index, obj], Commitment[num, _index, obj])
    # num is serialized 5bytes integer, other members have arbitrary size
    serialized_pruned_inputs=b""
    serialized_pruned_inputs+=len(self.pruned_inputs).to_bytes(2,"big")
    for _t, _c in self.pruned_inputs:
      for _i in [_t,_c]:
        num, _index, obj = _i
        serialized_pruned_inputs += num.to_bytes(5,"big") + self.serialize_bytes(_index) + self.serialize_bytes(obj)
    serialized_nums = self.num_of_added_outputs.to_bytes(2,"big") + self.num_of_added_excesses.to_bytes(2,"big")
    version=b"\x01"
    serialized_state_id = self.serialize_bytes(self.prev_state)
    summary_len = len(serialized_pruned_inputs)+len(serialized_nums)+len(version)+len(serialized_state_id)
    serialized = summary_len.to_bytes(4,"big")+version+serialized_pruned_inputs+serialized_nums+serialized_state_id
    return serialized

  def deserialize_raw(self, serialized):
    _slen, serialized = serialized[:4], serialized[4:]
    _len = int.from_bytes(_slen, "big")
    data, residue = serialized[:_len], serialized[_len:]
    version, data = data[:1], data[1:]
    if not version==b"\x01":
      raise
    _s_pruned_inputs_num, data = data[:2], data[2:]
    pruned_inputs_num = int.from_bytes(_s_pruned_inputs_num, "big")
    for i in range(pruned_inputs_num):
      _snum, data = data[:5], data[5:]
      _num = int.from_bytes(_snum, "big")
      _index, data = self.deserialize_bytes_raw(data)
      _obj, data = self.deserialize_bytes_raw(data)
      _txout = [_num,_index,_obj]
      _snum, data = data[:5], data[5:]
      _num = int.from_bytes(_snum, "big")
      _index, data = self.deserialize_bytes_raw(data)
      _obj, data = self.deserialize_bytes_raw(data)
      _comm = [_num,_index,_obj]
      self.pruned_inputs.append([_txout, _comm])
    _no, _ne, data = data[:2], data[2:4], data[4:]
    self.num_of_added_outputs = int.from_bytes(_no,"big")
    self.num_of_added_excesses = int.from_bytes(_ne, "big")
    self.prev_state, data = self.deserialize_bytes_raw(data)
    return residue


class BlocksDiscStorage:
  def __init__(self, dir_path):
    self.dir_path = dir_path

    if not os.path.exists(self.dir_path): 
        os.makedirs(self.dir_path) #TODO catch
    self.env = lmdb.open(self.dir_path, max_dbs=10)
    with self.env.begin(write=True) as txn:
      self.main_db = self.env.open_db(b'main_db', txn=txn, dupsort=False) # block_hash -> serialized_contextblock
      self.revert_db = self.env.open_db(b'revert_db', txn=txn, dupsort=False) # block_hash -> object_for_reverting

  def put(self, _hash, serialized_block):
    with self.env.begin(write=True) as txn:
      p1=txn.put( bytes(_hash), bytes(serialized_block), db=self.main_db, dupdata=False, overwrite=True)

  def update(self, _hash, serialized_block):
    with self.env.begin(write=True) as txn:
      txn.put(bytes(_hash), bytes(serialized_block), db=self.main_db, dupdata=False, overwrite=True)

  def get_by_hash(self, _hash):
    with self.env.begin(write=False) as txn:
      return txn.get(bytes(_hash), db=self.main_db)

  def get_rollback_object(self, _hash):
    with self.env.begin(write=False) as txn:
      return txn.get(bytes(_hash), db=self.revert_db)

  def pop_rollback_object(self, _hash):
    with self.env.begin(write=True) as txn:
      return txn.pop(bytes(_hash), db=self.revert_db)

  def put_rollback_object(self, _hash, serialized_rollback_object):
    with self.env.begin(write=True) as txn:
      return txn.put(bytes(_hash), bytes(serialized_rollback_object), db=self.revert_db)

  def __contains__(self, _hash):
    return bool(self.get_by_hash(_hash))

  def delete_block_by_hash(self, _hash):
    with self.env.begin(write=True) as txn:
      txn.delete( bytes(_hash), db=self.main_db)
    
  




from leer.core.storage.merkle_storage import MMR
from secp256k1_zkp import PedersenCommitment, PublicKey
from leer.core.storage.default_paths import txo_storage_path
from leer.core.lubbadubdub.ioput import IOput
from leer.core.utils import sha256
import os




class CommitmentMMR(MMR):
      def sum(self, x1,x2):
        # each index is 33 bytes for commitments and 32 for hash
        comm1, hash1 = x1[:33], x1[33:65]
        comm2, hash2 = x2[:33], x2[33:65]
        comm1, comm2 = PedersenCommitment(commitment=comm1, raw=True), PedersenCommitment(commitment=comm2, raw=True)
        #XXX we definetely need sum of pedersen commitments on libsecp256 level.
        pk= PublicKey()
        pk.combine([comm1.to_public_key().public_key, comm2.to_public_key().public_key])
        sm = pk.to_pedersen_commitment()
        first_part = sm.serialize()
        second_part = sha256(hash1+hash2)
        return first_part+second_part
        

class TXOMMR(MMR):
      def sum(self, x1,x2):
        return sha256(x1+x2)

class ConfirmedTXOStorage:
    '''
      Storage for all TXOs which are already in blocks.
      It is double MMR tree: first for commitments only, second for full txout
      Commitment tree:
        Each commitment leaf obj is `b""` and commitment leaf index is `bytes(serialized_apc)+bytes(hash(serialized_apc))`.
        Summ of two nodes in commitment tree = bytes(serialized_apc1+serialized_apc1)+hash(hash(serialized_apc)+hash(serialized_apc)),
        where commitments are summed as points on curve and summ of hashes is concantenation.

      Txout tree:
        Each txout leaf obj is serialized ioput, leaf index is hash(ioput_index).
        Summ of two nodes is hash of concantenation

    '''

    def __init__(self, path=txo_storage_path):
      self.commitments = CommitmentMMR("commitments", os.path.join(path, "confirmed"), clear_only=False)
      self.txos = TXOMMR("txos", os.path.join(path, "confirmed"), discard_only=True)

    def __getitem__(self, hash_and_pc):
      res = self.txos.get_by_hash(sha256(hash_and_pc))
      if not res:
        raise KeyError(hash_and_pc)
      utxo=IOput()
      utxo.deserialize(res)
      utxo.set_verified_and_correct() #Trust saved outputs
      return utxo

    #TODO Full method removement is postponed till comments inside will be resolved
    #def __setitem__(self, hash_and_pc, utxo):
    #  #TODO __setitem__ interface should be substituted with append-like interface
    #  #XXX problem here: commitments indexes should be apc+hash(apc), not hash_and_pc
    #  #here we should save
    #  self.txos.append(_hash(hash_and_pc),utxo.serialize())
    #  self.commitments.append(utxo.commitment_index,b"")

    def append(self, utxo):
      assert utxo.verify() #Should be fast due since cached
      self.txos.append(sha256(utxo.serialized_index), utxo.serialize())
      self.commitments.append(utxo.commitment_index,b"")

    def spend(self, utxo, return_revert_obj=False):
      txos = self.txos.discard(sha256(utxo.serialized_index))
      commitment = self.commitments.clear(utxo.commitment_index)
      if return_revert_obj:
        return (txos, commitment)

    def find(self, hash_and_pc):
      '''
        In contrast with __getitem__ find will try to find even spent
        outputs for other (syncing) nodes.
      '''
      res = self.txos.find_by_hash(sha256(hash_and_pc))
      if not res:
        raise KeyError(hash_and_pc)
      utxo=IOput()
      utxo.deserialize(res)
      return utxo

    def unspend(self, revert_obj):
      (txos, commitment) = revert_obj
      self.txos.revert_discarding(txos)
      self.commitments.revert_clearing(commitment)

    def __contains__(self, serialized_index):
      return bool(self.txos.get_by_hash(sha256(serialized_index)))

    def remove(self,n):
      self.commitments.remove(n)
      ser_removed_outputs = self.txos.remove(n)
      removed_outputs=[]
      for _ser in ser_removed_outputs:
        utxo=IOput()
        utxo.deserialize(_ser)
        removed_outputs.append(utxo)
      return removed_outputs

    def get_commitment_root(self):
      return self.commitments.get_root()

    def get_txo_root(self):
      return self.txos.get_root()

    def get_state(self):
      return self.commitments.get_state()

    def set_state(self, state):
      self.txos.set_state(state)
      self.commitments.set_state(state)

    def find_wo_deser(self, hash_and_pc):
      res = self.txos.find_by_hash(sha256(hash_and_pc))
      if not res:
        raise KeyError(hash_and_pc)
      return res


class TXOsStorage:

  class Interface:

    def __init__(self):
      self.storage = {}

    def __getitem__(self, hash_and_pc):
      if not hash_and_pc in self.storage:
        raise KeyError(hash_and_pc)
      return self.storage[hash_and_pc]

    def __setitem__(self, hash_and_pc, utxo):
      #TODO __setitem__ interface should be substituted with append-like interface
      #here we should save
      self.storage[hash_and_pc]=utxo

    def remove(self, utxo):
      self.remove_by_index(utxo.serialized_index)

    def remove_by_index(self, _index):
      self.storage.pop(_index)

    def __contains__(self, utxo):
      return utxo in self.storage  

    def flush(self):
      self.storage = {}


  __shared_states = {}


  def __init__(self, storage_space, path):
    if not path in self.__shared_states:
        self.__shared_states[path]={}
    self.__dict__ = self.__shared_states[path]
    self.path = path
    self.confirmed = ConfirmedTXOStorage(self.path)
    self.mempool = self.Interface()
    self.storage_space = storage_space
    self.storage_space.register_txos_storage(self)
      

  def known(self, output_index):
      return (output_index in self.confirmed) or (output_index in self.mempool)

  def confirm(self, output_index):
    utxo = self.mempool.storage.pop(output_index)
    self.confirmed.append(utxo)

  def apply_tx_get_merkles_and_rollback(self, tx):
    rollback_inputs = []
    for _i in tx.inputs:
        rollback_inputs.append(self.confirmed.spend(_i, return_revert_obj=True))
    for _o in tx.outputs:
        self.confirmed.append(_o)
    roots=[self.confirmed.get_commitment_root(), self.confirmed.get_txo_root()]
    for r_i in rollback_inputs:
      self.confirmed.unspend(r_i)
    self.confirmed.remove(len(tx.outputs))
    return roots

  #TODO bad naming. It should be apply block, or block_transaction
  def apply_tx(self, tx, new_state):
    rollback_inputs = []
    for _i in tx.inputs:
        if self.storage_space.utxo_index:
          self.storage_space.utxo_index.remove_utxo(_i)
        rollback_inputs.append(self.confirmed.spend(_i, return_revert_obj=True))
    for _o in tx.outputs:
        if self.storage_space.utxo_index:
          self.storage_space.utxo_index.add_utxo(_o)
        self.confirmed.append(_o)
        self.mempool.remove(_o)
    self.confirmed.set_state(new_state)
    return (rollback_inputs, len(tx.outputs))

  def rollback(self, pruned_inputs, num_of_added_outputs, prev_state):
    for r_i in pruned_inputs:
      if self.storage_space.utxo_index:
        #r_i[0][2] is serialized txo (0 is txo, 2 is serialized object)
        utxo=IOput()
        utxo.deserialize(r_i[0][2])
        self.storage_space.utxo_index.add_utxo(utxo)
      self.confirmed.unspend(r_i)
    outputs_for_mempool = self.confirmed.remove(num_of_added_outputs)
    for _o in outputs_for_mempool:
      self.mempool[_o.serialized_index]=_o
      self.storage_space.utxo_index.remove_utxo(_o)
    self.confirmed.set_state(prev_state)
        

  def find_serialized(self, output_index):
    if output_index in self.mempool:
      return self.mempool[output_index].serialize()
    else:
      return self.confirmed.find_wo_deser(output_index)
    





from leer.core.chains.headers_manager import HeadersManager
from leer.core.chains.blockchain import Blockchain
from leer.core.storage.txos_storage import TXOsStorage
from leer.core.storage.headers_storage import HeadersStorage
from leer.core.storage.blocks_storage import BlocksStorage
from leer.core.storage.excesses_storage import ExcessesStorage
from leer.core.storage.utxo_index_storage import UTXOIndex
from leer.core.storage.mempool_tx import MempoolTx
from leer.core.storage.key_manager import KeyManagerClass
from leer.core.primitives.block import Block
from leer.core.primitives.header import Header
from leer.core.lubbadubdub.ioput import IOput
from leer.core.lubbadubdub.address import Address
from leer.core.lubbadubdub.transaction import Transaction
import base64
from leer.core.utils import DOSException
from leer.core.primitives.transaction_skeleton import TransactionSkeleton
from time import sleep, time
from uuid import uuid4

from leer.core.storage.storage_space import StorageSpace
from leer.core.storage.default_paths import base_dir as default_base_dir, calc_paths

from leer.syncer import Syncer
from leer.core.parameters.constants import serialized_genesis_block

from secp256k1_zkp import PrivateKey

import logging
from functools import partial
logger = logging.getLogger("core_loop")

storage_space = StorageSpace()

def init_blockchain():
  '''
    If blockchain is empty this function will set genesis.
  '''
  genesis = Block(storage_space = storage_space)
  genesis.deserialize(serialized_genesis_block)
  storage_space.headers_manager.set_genesis(genesis.header)
  if not storage_space.blockchain.current_height>=0:
    storage_space.headers_manager.context_validation(genesis.header.hash)
    genesis.non_context_verify()
    storage_space.blockchain.add_block(genesis)
  else:
    storage_space.headers_manager.best_tip = (storage_space.blockchain.current_tip, storage_space.blockchain.current_height )
    logger.info("Best header tip from blockchain state %d"%storage_space.headers_manager.best_tip[1])
    #greedy search
    current_tip = storage_space.headers_manager.best_tip[0]
    while True:
      try:
        header = storage_space.headers_storage[current_tip]
      except KeyError:
        break
      new_current_tip=current_tip
      if len(header.descendants):
        for d in header.descendants:
          dh = storage_space.headers_storage[d]
          if not dh.invalid:
            new_current_tip = d
            break
      if not new_current_tip == current_tip:
        current_tip=new_current_tip
      else:
        break
    storage_space.headers_manager.best_tip = (current_tip, storage_space.headers_storage[current_tip].height)
    logger.info("Best header tip after greedy search %d"%storage_space.headers_manager.best_tip[1])


def init_storage_space(config):
  _paths = {}
  _paths["txo_storage_path"], _paths[ "txo_storage_path"], _paths[ "excesses_storage_path"],\
  _paths[ "headers_storage_path"], _paths[ "blocks_storage_path"], _paths[ "wallet_path"],\
  _paths[ "key_manager_path"], _paths[ "utxo_index_path"] = calc_paths(default_base_dir)
  if "location" in config:
    if "basedir" in config["location"]:
      basedir = config["location"]["basedir"]
      _paths["txo_storage_path"], _paths[ "txo_storage_path"],\
      _paths[ "excesses_storage_path"], _paths[ "headers_storage_path"],\
      _paths[ "blocks_storage_path"], _paths[ "wallet_path"],\
      _paths[ "key_manager_path"], _paths[ "utxo_index_path"] = calc_paths(basedir)
    for _path in _paths:
      if _path in config["location"]:
        _paths[_path] = config["location"][_path]
      
  hs = HeadersStorage(storage_space, _paths["headers_storage_path"])
  hm = HeadersManager(storage_space)
  bs = BlocksStorage(storage_space, _paths["blocks_storage_path"])
  es = ExcessesStorage(storage_space, _paths["excesses_storage_path"])
  ts = TXOsStorage(storage_space, _paths["txo_storage_path"])
  bc = Blockchain(storage_space)
  mptx = MempoolTx(storage_space)
  utxoi = UTXOIndex(storage_space, _paths["utxo_index_path"])
  km = KeyManagerClass(path = _paths["key_manager_path"]) #TODO km should be initialised in wallet process
  mptx.set_key_manager(km)
  init_blockchain()

  

def set_ask_for_blocks_hook(blockchain, message_queue):
  def f(block_hashes):
    if not isinstance(block_hashes, list):
      block_hashes=[block_hashes] #There is only one block
    new_message = {"action": "check blocks download status", "block_hashes":block_hashes,
                         "already_asked_nodes": [], "id": str(uuid4()),
                         "time": -1 }
    message_queue.put(new_message)
  
  blockchain.ask_for_blocks_hook = f

def set_ask_for_txouts_hook(block_storage, message_queue):
  def f(txouts):
    new_message = {"action": "check txouts download status", "txos_hashes": txouts,
                         "already_asked_nodes": [], "id": str(uuid4()),
                         "time": -1 }
    message_queue.put(new_message)
  
  block_storage.ask_for_txouts_hook = f

def set_notify_wallet_hook(blockchain, wallet_message_queue):
    def notify_wallet(reason, *args):
      message={'sender':"Blockchain"}
      #no id: notification
      if reason == "apply":
        message['action'] = "process new block"
        message['tx'] = args[0].serialize()
        message['height'] = args[1]      
      elif reason == "rollback":
        message['action'] = "process rollback"
        message['rollback_object'] = args[0].serialize()
        message['block_height'] = args[1]
      else:
        pass
      wallet_message_queue.put(message)
    storage_space.blockchain.notify_wallet = notify_wallet 

def core_loop(syncer, config):
  message_queue = syncer.queues['Blockchain']
  init_storage_space(config)    

  nodes = {}
  set_ask_for_blocks_hook(storage_space.blockchain, message_queue)
  set_ask_for_txouts_hook(storage_space.blocks_storage, message_queue)
  if config['wallet']:
    set_notify_wallet_hook(storage_space.blockchain, syncer.queues['Wallet'])
  requests = {}
  message_queue.put({"action":"give nodes list reminder"})

  def get_new_address(): #blocking
    _id = str(uuid4())
    syncer.queues['Wallet'].put({'action':'give new address', 'id':_id, 'sender': "Blockchain"})
    while True:
      put_back = [] #We wait for specific message, all others will wait for being processed
      while not message_queue.empty():
        message = message_queue.get()
        if (not 'id' in message)  or (not message['id']==_id):
          put_back.append(message)
          continue
        if message['result']=='error':
          raise KeyError
        address = Address()
        logger.info("Receiving address %s (len %d)"%( message["result"], len(message["result"])))

        address.deserialize_raw(message['result'])
        return address
      sleep(0.01)
      for message in put_back:
        message_queue.put(message)

  def send_message(destination, message):
    if not 'id' in message:
      message['id'] = uuid4()
    if not 'sender' in message:
      message['sender'] = "Blockchain"
    syncer.queues[destination].put(message)

  def send_to_nm(message):
    send_message("NetworkManager", message)

  notification_cache = {}
  def notify(key, value, timestamp=None):
    if (key in notification_cache) and (notification_cache[key]['value'] == value) and (time()-notification_cache[key]['timestamp'])<5:
      return #Do not spam notifications with the same values
    message = {}
    message['id'] = uuid4()
    message['sender'] = "Blockchain"
    if not timestamp:
      timestamp = time()
    message['time'] = timestamp
    message['action']="set"
    message['key']=key
    message['value']=value
    syncer.queues["Notifications"].put(message)
    notification_cache[key] = {'value':value, 'timestamp':timestamp}

  logger.debug("Start of core loop")
  while True:
    sleep(0.05)
    put_back_messages = []
    notify("core workload", "idle")
    while not message_queue.empty():
      message = message_queue.get()
      if 'time' in message and message['time']>time(): # delay this message
        put_back_messages.append(message)
        continue
      logger.info("Processing message %s"%message)
      if not 'action' in message: #it is response
        if message['id'] in requests: # response is awaited
          if requests[message['id']]=="give nodes list":
            requests.pop(message['id'])
            message_queue.put({"action":"take nodes list", "nodes":message["result"]})
        else:
          pass #Drop
        continue
      try:
        if message["action"] == "take the headers":
          notify("core workload", "processing new headers")
          process_new_headers(message)
          notify("best header", storage_space.headers_manager.best_header_height)         
        if message["action"] == "take the blocks":
          notify("core workload", "processing new blocks")
          initial_tip = storage_space.blockchain.current_tip
          process_new_blocks(message)
          after_tip = storage_space.blockchain.current_tip
          if not after_tip==initial_tip:
            notify_all_nodes_about_new_tip(nodes, send_to_nm) 
          notify("blockchain height", storage_space.blockchain.current_height)         
          look_forward(nodes, send_to_nm)       
        if message["action"] == "take the txos":
          notify("core workload", "processing new txos")
          process_new_txos(message)
          #After downloading new txos some blocs may become downloaded
          notify("blockchain height", storage_space.blockchain.current_height) 
          look_forward(nodes, send_to_nm)          
        if message["action"] == "give blocks":
          notify("core workload", "giving blocks")
          process_blocks_request(message, send_message)
        if message["action"] == "give next headers":
          notify("core workload", "giving headers")
          process_next_headers_request(message, send_message)
        if message["action"] == "give txos":
          notify("core workload", "giving txos")
          process_txos_request(message, send_message)
        if message["action"] == "find common root":
          process_find_common_root(message, send_message)
        if message["action"] == "find common root response":
          process_find_common_root_reponse(message, nodes[message["node"]], send_message)
        if message["action"] == "give TBM transaction":
          notify("core workload", "giving mempool tx")
          process_tbm_tx_request(message, send_message)
        if message["action"] == "take TBM transaction":
          notify("core workload", "processing mempool tx")
          process_tbm_tx(message, send_to_nm, nodes)
        if message["action"] == "give tip height":
          _ch=storage_space.blockchain.current_height
          send_message(message["sender"], {"id": message["id"], "result": _ch})
          notify("blockchain height", _ch)
      
        if message["action"] == "take tip info":
          if not message["node"] in nodes:
            nodes[message["node"]]={'node':message["node"]}
          process_tip_info(message, nodes[message["node"]], send=send_to_nm)
      except DOSException as e:
        logger.info("DOS Exception %s"%str(e))
        #raise e #TODO send to NM
      except Exception as e:
        raise e

      if message["action"] == "give block template":
        notify("core workload", "generating block template")
        address = get_new_address()
        block = storage_space.mempool_tx.give_block_template(address)
        ser_head = block.header.serialize()
        send_message(message["sender"], {"id": message["id"], "result":ser_head})
      if message["action"] == "take solved block template":
        notify("core workload", "processing solved block")
        try:
          initial_tip = storage_space.blockchain.current_tip
          header = Header()
          header.deserialize(message["solved template"])
          solved_block = storage_space.mempool_tx.get_block_by_header_solution(header)
          storage_space.headers_manager.add_header(solved_block.header)
          storage_space.headers_manager.context_validation(solved_block.header.hash)
          solved_block.non_context_verify()
          storage_space.blockchain.add_block(solved_block)
          send_message(message["sender"], {"id": message["id"], "result": "Accepted"})
          after_tip = storage_space.blockchain.current_tip
          if not after_tip==initial_tip:
            notify_all_nodes_about_new_tip(nodes, send_to_nm)
          our_height = storage_space.blockchain.current_height
          best_known_header = storage_space.headers_manager.best_header_height
          notify("best header", best_known_header)
          notify("blockchain height", our_height)
        except Exception as e:
          logger.ERROR("Wrong block solution %s"%str(e))
          send_message(message["sender"], {"id": message["id"], "error": str(e)})

      if message["action"] == "get confirmed balance stats": #TODO Move to wallet
        notify("core workload", "retrieving balance")
        if storage_space.mempool_tx.key_manager:
          stats = storage_space.mempool_tx.key_manager.get_confirmed_balance_stats( 
                     storage_space.utxo_index,
                     storage_space.txos_storage,
                     storage_space.blockchain.current_height)
          send_message(message["sender"], {"id": message["id"], "result":stats})
        else:
          send_message(message["sender"], {"id": message["id"], "error": "No registered key manager"})

      if message["action"] == "get confirmed balance list": #TODO Move to wallet
        notify("core workload", "retrieving balance")
        if storage_space.mempool_tx.key_manager:
          _list = storage_space.mempool_tx.key_manager.get_confirmed_balance_list( 
                     storage_space.utxo_index,
                     storage_space.txos_storage,
                     storage_space.blockchain.current_height)
          send_message(message["sender"], {"id": message["id"], "result":_list})
        else:
          send_message(message["sender"], {"id": message["id"], "error": "No registered key manager"})

      if message["action"] == "give new address": #TODO Move to wallet
        notify("core workload", "retrieving new address")
        if storage_space.mempool_tx.key_manager:
          texted_address = storage_space.mempool_tx.key_manager.new_address().to_text()
          send_message(message["sender"], {"id": message["id"], "result": texted_address})
        else:
          send_message(message["sender"], {"id": message["id"], "error": "No registered key manager"})

      if message["action"] == "give private key": #TODO Move to wallet
        if storage_space.mempool_tx.key_manager:
          km = storage_space.mempool_tx.key_manager
          a=Address()
          a.from_text(message["address"])
          serialized_pk = km.priv_by_address(a).serialize()
          send_message(message["sender"], {"id": message["id"], "result": serialized_pk})
        else:
          send_message(message["sender"], {"id": message["id"], "error": "No registered key manager"})

      if message["action"] == "take private key": #TODO Move to wallet
        if storage_space.mempool_tx.key_manager:
          km = storage_space.mempool_tx.key_manager
          pk=PrivateKey()
          pk.deserialize(message['privkey'])
          km.add_privkey(pk)
          send_message(message["sender"], {"id": message["id"], "result": "imported"})
        else:
          send_message(message["sender"], {"id": message["id"], "error": "No registered key manager"})

      if message["action"] == "give synchronization status":
        our_height = storage_space.blockchain.current_height
        best_known_header = storage_space.headers_manager.best_header_height
        try:
          best_advertised_height = max([nodes[node]["height"] for node in nodes if "height" in nodes[node]])
        except:
          best_advertised_height = None
        send_message(message["sender"], {"id": message["id"], 
                                         "result": {'height': our_height, 
                                                    'best_known_header': best_known_header,
                                                    'best_advertised_height': best_advertised_height}})
        notify("best header", best_known_header)
        notify("blockchain height", our_height)
        notify("best advertised height", best_advertised_height)


      if message["action"] == "generate tx by tx template": #TODO Move to wallet
        notify("core workload", "generating transactions")
        response = {"id": message["id"]}
        try:
          tx_template = message["tx_template"]
          #Deserialization        
          destination_address, change_address = Address(), Address()
          destination_address.deserialize_raw(tx_template['address'])
          change_address.deserialize_raw(tx_template['change address'])
          tx_template['address'], tx_template['change address'] = destination_address, change_address
          for pub in tx_template['priv_by_pub']:
            tx_template['priv_by_pub'][pub] =  PrivateKey(tx_template['priv_by_pub'][pub], raw=True)
        except Exception as e:
          response['result'] = 'error'
          response['error'] = str(e)
          logger.error("Problem in tx_template: %s"%str(e))
        try: #Tx generation
          tx = Transaction(txos_storage = storage_space.txos_storage)
          for utxo_index in tx_template['utxos']:
            utxo = storage_space.txos_storage.confirmed[utxo_index]
            tx.push_input(utxo)
          tx.add_destination( (tx_template["address"], tx_template["value"]) )
          tx.generate_new(priv_data=tx_template,
                        change_address = tx_template['change address'],
                        relay_fee_per_kb=storage_space.mempool_tx.fee_policy_checker.relay_fee_per_kb)
          tx.verify()
          storage_space.mempool_tx.add_tx(tx)
          tx_skel = TransactionSkeleton(tx=tx)
          notify_all_nodes_about_tx(tx_skel.serialize(rich_format=True, max_size=40000), nodes, send_to_nm, _except=[], mode=1)
          response['result']="generated"
        except Exception as e:
          response['result'] = 'error'
          response['error'] = str(e)
          logger.error("Cannot generate tx by template: %s"%str(e))
        send_message(message["sender"], response)

      #message from core_loop
      if message["action"] == "check txouts download status":
        txos = message["txos_hashes"]
        to_be_downloaded = []
        for txo in txos:
          if not storage_space.txos_storage.known(txo):
            to_be_downloaded.append(txo)
        if not to_be_downloaded:
          continue #We are good, txouts are already downloaded
        already_asked_nodes = message["already_asked_nodes"]
        asked = False
        for node_params in nodes:
          node = nodes[node_params]
          if node in already_asked_nodes:
            continue
          already_asked_nodes += [node]
          send_to_nm({"action":"give txos",
                                               "txos_hashes": b"".join(to_be_downloaded), 
                                               "num": len(to_be_downloaded), 
                                               "id":str(uuid4()), "node":node_params })
          new_message = {"action": "check txouts download status", "txos_hashes":to_be_downloaded,
                         "already_asked_nodes": already_asked_nodes, "id": str(uuid4()),
                         "time": int(time()+300) }
          asked = True
          put_back_messages.append(new_message)
          break
        if not asked: #We already asked all applicable nodes
          message["time"]=int(time())+3600
          message["already_asked_nodes"] = []
          put_back_messages.append(message) # we will try to ask again in an hour

      #message from core_loop
      if message["action"] == "check blocks download status":
        #TODO download many blocks at once
        block_hashes = message["block_hashes"]
        to_be_downloaded = []
        lowest_height=1e10
        for block_hash in block_hashes:
          if block_hash in storage_space.blocks_storage:
            continue #We are good, block already downloaded          
          if not block_hash in storage_space.blockchain.awaited_blocks:
            continue #For some reason we don't need this block anymore
          to_be_downloaded.append(block_hash)
          if storage_space.headers_storage[block_hash].height<lowest_height:
            lowest_height = storage_space.headers_storage[block_hash].height
        already_asked_nodes = message["already_asked_nodes"]
        asked = False
        for node_params in nodes:
          node = nodes[node_params]
          if node in already_asked_nodes:
            continue
          if (not "height" in node) or node["height"] < lowest_height:
            continue
          already_asked_nodes += [node]
          send_to_nm({"action":"give blocks",  "block_hashes": bytes(b"".join(block_hashes)), 'num': len(block_hashes), "id":str(uuid4()), "node":node_params })
          new_message = {"action": "check blocks download status", "block_hashes":to_be_downloaded,
                         "already_asked_nodes": already_asked_nodes, "id": str(uuid4()),
                         "time": int(time()+300) }
          asked = True
          put_back_messages.append(new_message)
          break
        if not asked: #We already asked all applicable nodes
          message["time"]=int(time())+3600
          message["already_asked_nodes"] = []
          put_back_messages.append(message) # we will try to ask again in an hour


      if message["action"] == "take nodes list":
        for node in message["nodes"]:
          if not node in nodes: #Do not overwrite
            nodes[node]={"node":node}
        disconnected_nodes = []
        for existing_node in nodes:
          if not existing_node in message["nodes"]:
            disconnected_nodes.append(existing_node)
        for dn in disconnected_nodes:
          nodes.pop(dn)
        

      if message["action"] == "give nodes list reminder":
        _id = str(uuid4())
        send_to_nm({"action":"give nodes list", "sender":"Blockchain", "id":_id})
        requests[_id] = "give nodes list"
        put_back_messages.append({"action": "give nodes list reminder", "time":int(time())+3} )

      if message["action"] == "stop":
        logger.info("Core loop stops")
        return

    for _message in put_back_messages:
      message_queue.put(_message)

    try:
      check_sync_status(nodes, send_to_nm)
      try:
        best_advertised_height = max([nodes[node]["height"] for node in nodes if "height" in nodes[node]])
      except:
          best_advertised_height = None
      notify("best advertised height", best_advertised_height)
    except Exception as e:
      logger.error(e)

def look_forward(nodes, send_to_nm):
  if storage_space.headers_manager.best_header_height < storage_space.blockchain.current_height+100:
    for node_index in nodes:
      node = nodes[node_index]
      if node['height']>storage_space.headers_manager.best_header_height:
        our_tip_hash = storage_space.blockchain.current_tip
        send_find_common_root(storage_space.headers_storage[our_tip_hash], node['node'], send = send_to_nm)
        break




def process_new_headers(message):
  dupplication_header_dos = False
  try:
    serialized_headers = message["headers"]
    num = message["num"]
    for i in range(num):
      header = Header()
      serialized_headers = header.deserialize_raw(serialized_headers)
      if header.hash in storage_space.headers_storage:
        dupplication_header_dos=True
        continue
      storage_space.headers_manager.add_header(header)
    storage_space.blockchain.update(reason="downloaded new headers")
  except Exception as e:
    raise e

def process_new_blocks(message):
  try:
    serialized_blocks = message["blocks"]
    num = message["num"]
    for i in range(num):
      block = Block(storage_space=storage_space)
      serialized_blocks = block.deserialize_raw(serialized_blocks)
      storage_space.blockchain.add_block(block, no_update=True)
    storage_space.blockchain.update(reason="downloaded new blocks")
  except Exception as e:
    raise e #XXX "DoS messages should be returned"

def process_new_txos(message):
  try:
    serialized_utxos = bytes(message["txos"])
    txos_hashes = bytes(message["txos_hashes"])
    num = message["num"]
    #txos_hashes = [txos_hashes[i*65:(i+1)*65)] for i in range(0,num)]
    txos_lengths = message["txos_lengths"]
    #TODO we should use txos_hashes and txos_lengths to proccess only unknown txos
    for i in range(num):
      utxo = IOput()
      serialized_utxos = utxo.deserialize_raw(serialized_utxos)
      storage_space.txos_storage.mempool[utxo.serialized_index]=utxo
    storage_space.blockchain.update(reason="downloaded new txos")
  except Exception as e:
    raise e #XXX "DoS messages should be returned"

def process_blocks_request(message, send_message):
  num = message["num"]
  _hashes = message["block_hashes"]
  _hashes = [_hashes[i*32:(i+1)*32] for i in range(num)]
  serialized_blocks = b""
  blocks_num=0
  for _hash in _hashes:
    if not _hash in storage_space.blocks_storage:
      continue
    try:
      serialized_block = storage_space.blocks_storage[_hash].serialize(rich_block_format=True)
    except KeyError:
      #Some outputs were pruned
      serialized_block = storage_space.blocks_storage[_hash].serialize(rich_block_format=False)
    if len(serialized_blocks)+len(serialized_block)<60000:
      serialized_blocks+=serialized_block
      blocks_num +=1
    else:
        send_message(message['sender'], {"action":"take the blocks", "num":blocks_num, 
                                              "blocks":serialized_blocks, "id":message['id'],
                                              "node":message["node"]})
        serialized_blocks=serialized_block
        blocks_num =1
  send_message(message['sender'], {"action":"take the blocks", "num":blocks_num, 
                                              "blocks":serialized_blocks, "id":message['id'],
                                              "node":message["node"]})


def send_next_headers_request(from_hash, num, node, send):
  send({"action":"give next headers", "num":num, "from":from_hash, 
                                       "id" : str(uuid4()), "node": node  })

def process_next_headers_request(message, send_message):
  from_hash = message["from"]
  num = message["num"]
  num = 1024 if num>1024 else num
  try:
    header = storage_space.headers_storage[from_hash]
  except KeyError:
    return #unknown hash
  current_tip  = storage_space.blockchain.current_tip
  current_height  = storage_space.blockchain.current_height
  if not storage_space.headers_manager.find_ancestor_with_height(current_tip, header.height) == from_hash:
    return
    ''' Counter-node is not in our main chain. 
        We will not feed it (actually we just are not sure what we should send here)
    '''
  last_to_send_height = header.height+num
  last_to_send_height = current_height if last_to_send_height>current_height else last_to_send_height
  last_to_send = storage_space.headers_manager.find_ancestor_with_height(current_tip, last_to_send_height)
  headers_hashes = storage_space.headers_manager.get_subchain(from_hash, last_to_send)

  serialized_headers = b""
  headers_num=0
  for _hash in headers_hashes:
    if not _hash in storage_space.headers_storage:
      continue
    serialized_header = storage_space.headers_storage[_hash].serialize()
    if len(serialized_headers)+len(serialized_header)<60000:
      serialized_headers+=serialized_header
      headers_num +=1
    else:
      send_message(message['sender'], {"action":"take the headers", "num": headers_num, 
                                            "headers":serialized_headers, "id":message['id'],
                                            "node": message["node"] })
      serialized_headers=serialized_header
      headers_num =1
  if headers_num:
    send_message(message['sender'], {"action":"take the headers", "num": headers_num, 
                                            "headers":serialized_headers, "id":message['id'],
                                            "node": message["node"] })

def process_txos_request(message, send_message):
  num = message["num"]
  _hashes = message["txos_hashes"]
  _hashes = [bytes(_hashes[i*65:(i+1)*65]) for i in range(num)]
  serialized_txos = b""
  txos_num=0
  txos_hashes = b""
  txos_lengths = b""
  for _hash in _hashes:
    try:
      serialized_txo = storage_space.txos_storage.find_serialized(_hash)
    except KeyError:
      continue
    if len(serialized_txos)+len(serialized_txo)<60000:
      serialized_txos+=serialized_txo
      txos_num +=1
      txos_hashes += _hash
      txos_lengths += len(serialized_txo).to_bytes(2,"big")
    else:
      send_message(message['sender'], {"action":"take the txos", 
                                            "num":txos_num, "txos":serialized_txos, 
                                            "txos_hashes": txos_hashes, "txos_lengths": txos_lengths,
                                            "id":message['id'], 'node': message["node"] })
      serialized_txos=serialized_txo
      txos_num =1
  if txos_num:
    send_message(message['sender'], {"action":"take the txos", 
                                            "num":txos_num, "txos":serialized_txos, 
                                            "txos_hashes": txos_hashes, "txos_lengths": txos_lengths,
                                            "id":message['id'], 'node': message["node"] })

def send_tip_info(node_info, send, our_tip_hash=None ):
  our_height = storage_space.blockchain.current_height
  our_tip_hash = our_tip_hash if our_tip_hash else storage_space.blockchain.current_tip
  our_prev_hash = storage_space.headers_storage[our_tip_hash].prev
  our_td = storage_space.headers_storage[our_tip_hash].total_difficulty

  send({"action":"take tip info", "height":our_height, "tip":our_tip_hash, "prev_hash":our_prev_hash, "total_difficulty":our_td, "id":uuid4(), "node": node_info["node"] })
  node_info["sent_tip"]=our_tip_hash
  node_info["last_send"] = time()

def process_tip_info(message, node_info, send):
  # another node (referenced as counter-Node below) asks us for our best tip and also provides us information about his
  node = message["node"]
  height = message["height"]
  tip_hash = message["tip"]
  prev_hash = message["prev_hash"]
  total_difficulty = message["total_difficulty"]

  our_tip_hash = storage_space.blockchain.current_tip

  if (not "sent_tip" in node_info) or (not node_info["sent_tip"]==our_tip_hash):
    send_tip_info(node_info=node_info, send = send, our_tip_hash=our_tip_hash)
  node_info.update({"node":node, "height":height, "tip_hash":tip_hash, 
                    "prev_hash":prev_hash, "total_difficulty":total_difficulty, 
                    "last_update":time()})
  if (height > storage_space.blockchain.current_height) and (total_difficulty > storage_space.headers_storage[our_tip_hash].total_difficulty):
    #Now there are two options: better headers are unknown or headers are known, but blocks are unknown or bad
    if (not tip_hash in storage_space.headers_storage):
      send_find_common_root(storage_space.headers_storage[our_tip_hash], node, send = send)
      #TODO check prev hash first
    else: #header is known
      header = storage_space.headers_storage[tip_hash]
      #TODO do we need to check for connection to genesis?
      common_root =  storage_space.headers_manager.find_bifurcation_point(tip_hash, our_tip_hash)
      if header.invalid:
        return #Nothing interesting, counter-node is on wrong chain
      if tip_hash in storage_space.blocks_storage:
        if storage_space.blocks_storage[tip_hash].invalid:
          return #Nothing interesting, counter-node is on wrong chain
      #download blocks
      blocks_to_download = []
      print("Here")
      for _block_hash in storage_space.headers_manager.get_subchain(common_root, tip_hash):
        print(_block_hash, _block_hash in storage_space.blocks_storage)
        if not _block_hash in storage_space.blocks_storage:
          blocks_to_download.append(_block_hash)
        else:
          storage_space.blocks_storage.is_block_downloaded(_block_hash)
        if len(blocks_to_download)*32>40000: #Too big for one message
          break
      if len(blocks_to_download):
        send({"action":"give blocks",  
            "block_hashes": b"".join(blocks_to_download),
            'num': len(blocks_to_download), "id":str(uuid4()), "node":message["node"] })
      

def send_find_common_root(from_header, node, send):
  send(
    {"action":"find common root", "serialized_header": from_header.serialize(), 
     "id":str(uuid4()), 
     "node": node })


UNKNOWN, INFORK, MAINCHAIN, ISOLATED = 0, 1, 2, 3

def  process_find_common_root(message, send_message):
  serialized_header = message["serialized_header"]
  header = Header()
  header.deserialize_raw(serialized_header)
  result = []
  for pointer in [header.hash]+header.popow.pointers:
    if not pointer in storage_space.headers_storage:
      result.append(UNKNOWN)
      continue
    ph = storage_space.headers_storage[pointer]
    if not ph.connected_to_genesis:
      result.append(ISOLATED)
      continue
    if storage_space.headers_manager.find_ancestor_with_height(storage_space.blockchain.current_tip, ph.height) == pointer:
      result.append(MAINCHAIN)
      continue
    result.append(INFORK)

  send_message(message['sender'], \
     {"action":"find common root response", "header_hash":header.hash,
      "flags_num": len(result), 
      "known_headers": b"".join([i.to_bytes(1,"big") for i in result]), 
      "id":message['id'], "node": message["node"] })

def process_find_common_root_reponse(message, node_info, send_message):
  header_hash = message["header_hash"]
  result = [int(i) for i in message["known_headers"]]
  header = storage_space.headers_storage[header_hash]
  root_found = False
  if not "common_root" in node_info:
    node_info["common_root"]={}

  for index, pointer in enumerate([header.hash]+header.popow.pointers):
    if result[index] in [MAINCHAIN, INFORK]:
        node_info["common_root"]["best_mutual"]=pointer
        best_mutual_height = storage_space.headers_storage[node_info["common_root"]["best_mutual"]].height
        break
    else:
      node_info["common_root"]["worst_nonmutual"]=pointer
  logger.info("Processing of fcrr")
  logger.info(node_info)
  if (not "worst_nonmutual" in node_info["common_root"]):
    #we are behind
    node_info["common_root"]["root"] = header_hash
    root_found = True
  if (not "best_mutual" in node_info["common_root"]):
    # genesis should always be mutual
    return
  logger.info(node_info)
  if not root_found:
    h1,h2 = storage_space.headers_storage[node_info["common_root"]["worst_nonmutual"]].height, storage_space.headers_storage[node_info["common_root"]["best_mutual"]].height
    if h1==h2+1:
      root_found = True
      node_info["common_root"]["root"] = node_info["common_root"]["best_mutual"]
    else:
      send_find_common_root(storage_space.headers_storage[node_info["common_root"]["worst_nonmutual"]], message['node'],\
                          send = partial(send_message, "NetworkManager") )
  logger.info(node_info)
  height, total_difficulty = node_info['height'],node_info['total_difficulty']
  logger.info((height, storage_space.headers_manager.best_header_height, total_difficulty , storage_space.headers_manager.best_header_total_difficulty))
  if root_found:
    if (height > storage_space.headers_manager.best_header_height) and (total_difficulty > storage_space.headers_manager.best_header_total_difficulty):
      send_next_headers_request(node_info["common_root"]["root"], 
                                min(256, height-storage_space.headers_storage[node_info["common_root"]["root"]].height),
                                message["node"], send = partial(send_message, "NetworkManager") )



def  process_tbm_tx_request(message, send_message):
  tx_skel = storage_space.mempool_tx.give_tx_skeleton()
  tx = storage_space.mempool_tx.give_tx()
  serialized_tx_skel = tx_skel.serialize(rich_format=True, max_size=60000, full_tx=tx)
  send_message(message['sender'], \
     {"action":"take TBM transaction", "tx_skel": serialized_tx_skel, "mode": 0,
      "id":message['id'], 'node': message["node"] })

def  process_tbm_tx(message, send, nodes):
  try:
    initial_tbm = storage_space.mempool_tx.give_tx()
    tx_skel = TransactionSkeleton()
    tx_skel.deserialize_raw(message['tx_skel'], storage_space = storage_space)
    storage_space.mempool_tx.add_tx(tx_skel)
    final_tbm = storage_space.mempool_tx.give_tx()
    if not message["mode"]==0: #If 0 it is response to our request
      if (not initial_tbm) or (not str(initial_tbm.serialize())==str(final_tbm.serialize())):
        notify_all_nodes_about_tx(message['tx_skel'], nodes, send, _except=[message["node"]])
  except Exception as e:
    print(e)
    pass

  
def check_sync_status(nodes, send):
  for node_index in nodes:
    node = nodes[node_index]
    if ((not "last_update" in node) or node["last_update"]+300<time()) and ((not "last_send" in node) or node["last_send"]+5<time()):
      #logger.info("\n node last_update %d was %.4f sec\n"%(("last_update" in node),   (time()-node["last_update"] if ("last_update" in node) else 0 )))
      send_tip_info(node_info = node, send=send)


def notify_all_nodes_about_tx(tx_skel, nodes, send, _except=[], mode=1):
  #TODO we should not notify about tx with low relay fee
  for node_index in nodes:
    if node_index in _except:
      continue
    node = nodes[node_index]
    send({"action":"take TBM transaction", "tx_skel": tx_skel, "mode": mode,
      "id":str(uuid4()), 'node': node["node"] })

def notify_all_nodes_about_new_tip(nodes, send):
  for node_index in nodes:
    node = nodes[node_index]
    send_tip_info(node_info=node, send=send)




from secp256k1_zkp import GeneratorOnCurve

default_generator = GeneratorOnCurve()
default_generator._from_seed(b'\x03\x07\x11'+b'\xfa'*29)
default_generator_ser=default_generator.serialize()

generators = { default_generator_ser: default_generator}

GLOBAL_TEST = {'utxo set': True, 
'key manager': True,
'lock_height checking':True, 'block_version checking': True, 'inputs index': True,
'skip combined excesses':True, 'skip additional excesses':False, 'spend from mempool':True}



block_time = 60 #in seconds
max_target_increase = 1.2
max_target_decrease = 0.8
initial_target = 2**240

COIN = 100000000
initial_reward = 600*COIN
reward_decrease_halflife = 2100000 # in blocks

output_creation_fee = 22000

coinbase_maturity = 60

serialized_genesis_block = b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\x08\xcd\xda\xcf\xa0X\x81\xdd\xfb\n\x9e\x10M\xf3!\xd6\xb5\xb1\x1eU\xa6\x98\xc54,1\xab\xa4\xaa\xf8SeU\xe1te\x90o\xde\xa3:-\xf8\xfa\xe8=\xe5\xdf\x8e\x84m!^\xa3$D\x89\xe2\x91\xa0k\xee\x01\xaef\xcd5^\xac\x15\x99\x1aJ\xfc\x06f\xbcE\x86\x00\x9b\xd2H\xc3\xb1\xfd\x01\\\xca6[_tPh\xdd\xc7\x03q\x11\xb0\xa4\xc0\xbfVE\xbaK\xb9EQS\x00j\x0e\xad!\x08\x13\x1d\xb7F\xda\x80\x81\xa8~I\n\xce\x99\\-\xdaD\xa2\xcf-\xfd3\x01\x7f\x0f^i\x08\x7f5\xe0\x83\x95\xe7\xaeQ\xe3\xcf\xf2s\xdfg\x96W\x00\x00\x00\r\xf8GX\x00\x00[\xff\x97\x00\x00\xf0\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01z\xa9\x03\x00\x00\x00\x01\x00\x01\x08\xcd\xda\xcf\xa0X\x81\xdd\xfb\n\x9e\x10M\xf3!\xd6\xb5\xb1\x1eU\xa6\x98\xc54,1\xab\xa4\xaa\xf8SeU\x8c\xf2\x9c\xf5\xf5p\xbf\x0b\x95Y\xea\xe9S}\x9el\x18\xb8\x17\xddN\x8eY=\x97d\xb4\x98\xa5\xe5(\x03\x00\x00\x00\x00\x80A\xe9\xf6\xf7\xc0\xfb\x0c\xe6\xdcE\xb3B\xa2%ce\xfd\x99\r\xbf\x9d\x1fA\xf7ND\xcf\x92^\x81\xf6\xf75O\xf1\xe8x\xf7\xe1\x7f\xc6\xc3@\xd5\x946`\xfc\xe2\xf0~E)\xd8Y\x0c\x19\xf9Jd\xe9T\xda\xe6\x00!\x08\xcd\xda\xcf\xa0X\x81\xdd\xfb\n\x9e\x10M\xf3!\xd6\xb5\xb1\x1eU\xa6\x98\xc54,1\xab\xa4\xaa\xf8SeU\x00\x01\x00\x00\x00\x01\x00\x00\x00<\x0bN\x13\x90\xb0bl\xf8j\xc8/q\xc4!\x06\x1c;\x19\xf2\xd3\xd0\xdc\xb1\x97u\xc6&\xb5\xc7]/\xda\xa0\x00\x00\x00\x00\x00\x00\x00\x00\x08\xcd\xda\xcf\xa0X\x81\xdd\xfb\n\x9e\x10M\xf3!\xd6\xb5\xb1\x1eU\xa6\x98\xc54,1\xab\xa4\xaa\xf8SeU\x019\xbe\xd17\xbevE\xa0\x8a\x00\xed=\xbb\xe6\xe6\x88|\xb0\xd2\x95\x80\xa2\xa4\x0e\xf0\xb9\xc5g\xb9\xd1 \x88#2\xaa\xb0\xc0\xa7;\x99\x07u\xdbu\xa0\x8a\x1a,\x9bn\x01\xdbV/\x8a\xbb\xa8]a\xd9\x8d\xee\xaf\xc9\x00Y\x03\xa2\xf9\x88z\xe2s\xd7~\xb97\xe5\xbd\xd3a\xa0\xde/!\xc1\x1f\x1a?\x08\xdb\xf4f\xe3`6B\x17.\xff\xf36\xc3\x1f\xcd\xb8\x13\xf3\xb4\x00\x9e\xefU\xe7g\xa5N\r\x8fl\x1e\x0b\xf5\x98\xa3\x1bE\xb4\xb3\xb0\xac\n\x836\xa7\x94]8\rq\x82\x9f\x93\xbdf\xc4s.\x0b\x7f\xfcz>\xb5\xd0\x00I \x00\x00\x00\r\xf8GX\x00j\xf1\xdb\xeb\x05?\x0bb\xb5\n\x03vk\xcc\xb6\xedekwF\xfak\x8d\x110\xd20\x12\x89\x1b>\xc4\x8d\x15grp?\x83\xc8\xe2k/q\\\x9c\x80O\x8fT;\xe7\x14\xf5\x91Ktg\xa5s\x13_\x8a\r"



from collections import OrderedDict
import struct

from secp256k1_zkp import PrivateKey, PublicKey, PedersenCommitment, RangeProof

from leer.core.lubbadubdub.constants import default_generator, default_generator_ser, generators, GLOBAL_TEST
from leer.core.lubbadubdub.address import Address, Excess, excess_from_private_key
from leer.core.lubbadubdub.ioput import IOput
from leer.core.storage.txos_storage import TXOsStorage


from leer.core.parameters.constants import coinbase_maturity, output_creation_fee
from leer.core.primitives.transaction_skeleton import TransactionSkeleton
from leer.core.storage.verification_cache import verification_cache

def is_sorted(lst, key=lambda x: x):
    for i, el in enumerate(lst[1:]):
        if key(el) < key(lst[i]): # i is the index of the previous element
            return False
    return True



class Transaction:

  def __init__(self, txos_storage, raw_tx=None, key_manager=None):
    #serializable data 
    self.inputs = []
    self.outputs = []
    self.additional_excesses = []
    self.combined_excesses = OrderedDict()
    #inner data
    self._destinations = [] #destinations are unprepared outputs
    self.coinbase = None
    self.new_outputs_fee = None
    self.txid = None
    self.serialized = None

    self.txos_storage = txos_storage #we can switch fron global txos_storage to stubs for testing
    self.key_manager = key_manager
    
    if raw_tx:
      self.deserialize(raw_tx)

  def push_input(self, _input):
    self.inputs.append(_input)

  def add_destination(self, destination):
    self._destinations.append(destination)


  #TODO coinbase should be @property

  def add_coinbase(self, coinbase_output):
    self.serialized = None
    self.coinbase = coinbase_output

  def compose_block_transaction(self, combined_transaction=None):
    self.serialized = None
    if not self.coinbase:
      raise Exception("coinbase output is required")
    cb = self.coinbase
    self.__init__(txos_storage = self.txos_storage) #reset self
    self.outputs = [cb]
    self.additional_excesses = [excess_from_private_key(cb.blinding_key, cb.serialized_index[:33])]
    if combined_transaction:
      new_tx = self.merge(combined_transaction)
      self.inputs = new_tx.inputs
      self.outputs = new_tx.outputs
      self.additional_excesses = new_tx.additional_excesses
      self.combined_excesses = new_tx.combined_excesses
    self.verify()


  # should be moved to wallet???
  def generate(self, change_address=None, relay_fee_per_kb=0): #TODO key_manager should be substituted with inputs_info = {..., 'new_address': '', 'priv_by_pub': {'':''}}
    self.serialized = None
    if self.coinbase:
      raise Exception("generate() can be used only for common transaction, to create block transaction as miner use compose_block_transaction")
    if not len(self.inputs):
      raise Exception("Tx should have at least one input")
    if not len(self._destinations):
      raise Exception("Tx should have at least one destination")
    for ioput in self.inputs:
      if not self.key_manager:
        raise Exception("Trying to generate tx which spends unknown input (KeyManager is None)")
      if not ioput.detect_value(self.key_manager):
        raise Exception("Trying to generate tx which spends unknown input")
    in_value = sum([ioput.value for ioput in self.inputs]) 
    out_value = sum([destination[1] for destination in self._destinations])
    relay_fee = self.calc_relay_fee(relay_fee_per_kb=relay_fee_per_kb)
    # +1 for destination is for change address
    self.fee = relay_fee + self.calc_new_outputs_fee(len(self.inputs), len(self._destinations)+1)
    remainder = in_value - out_value - self.fee
    if remainder<0:
      raise Exception("Not enough money in inputs to cover outputs")
    # TODO We need logic here to cover too low remainders (less than new output fee)
    change_address =  change_address if change_address else self.key_manager.new_address() # TODO Check: for first glance, we cannot get here if key_manager is None.
    self._destinations.append((change_address, remainder))
    privkey_sum=0
    out_blinding_key_sum = None
    for out_index in range(len(self._destinations)-1):
      address, value = self._destinations[out_index]
      output = IOput()
      output.fill(address, value, generator = default_generator_ser)
      self.outputs.append( output )
      out_blinding_key_sum = out_blinding_key_sum + output.blinding_key if out_blinding_key_sum else output.blinding_key
    # privkey for the last one output isn't arbitrary
    address, value = self._destinations[-1]
    in_blinding_key_sum = None
    for _input in self.inputs:
      in_blinding_key_sum = in_blinding_key_sum + _input.blinding_key if in_blinding_key_sum else _input.blinding_key
      in_blinding_key_sum += self.key_manager.priv_by_pub(_input.address.pubkey) # we can't get here if key_manager is None
    output = IOput()
    output.fill(address, value, blinding_key = in_blinding_key_sum-out_blinding_key_sum,
      relay_fee=relay_fee, generator = default_generator_ser) #TODO relay fee should be distributed uniformly, privacy leak
    self.outputs.append(output)
    [output.generate() for output in self.outputs]
    self.sort_ioputs()
    self.verify()

  # should be moved to wallet???
  def generate_new(self, priv_data, change_address=None, relay_fee_per_kb=0): #TODO key_manager should be substituted with inputs_info = {..., 'new_address': '', 'priv_by_pub': {'':''}}
    self.serialized = None
    if self.coinbase:
      raise Exception("generate() can be used only for common transaction, to create block transaction as miner use compose_block_transaction")
    if not len(self.inputs):
      raise Exception("Tx should have at least one input")
    if not len(self._destinations):
      raise Exception("Tx should have at least one destination")
    for ioput in self.inputs:
      if not ioput.detect_value_new(inputs_info=priv_data):
        raise Exception("Trying to generate tx which spends unknown input")
    in_value = sum([ioput.value for ioput in self.inputs]) 
    out_value = sum([destination[1] for destination in self._destinations])
    relay_fee = self.calc_relay_fee(relay_fee_per_kb=relay_fee_per_kb)
    # +1 for destination is for change address
    self.fee = relay_fee + self.calc_new_outputs_fee(len(self.inputs), len(self._destinations)+1)
    remainder = in_value - out_value - self.fee
    if remainder<0:
      raise Exception("Not enough money in inputs to cover outputs")
    # TODO We need logic here to cover too low remainders (less than new output fee)
    change_address =  change_address if change_address else priv_data['change address']
    self._destinations.append((change_address, remainder))
    privkey_sum=0
    out_blinding_key_sum = None
    for out_index in range(len(self._destinations)-1):
      address, value = self._destinations[out_index]
      output = IOput()
      output.fill(address, value, generator = default_generator_ser)
      self.outputs.append( output )
      out_blinding_key_sum = out_blinding_key_sum + output.blinding_key if out_blinding_key_sum else output.blinding_key
    # privkey for the last one output isn't arbitrary
    address, value = self._destinations[-1]
    in_blinding_key_sum = None
    for _input in self.inputs:
      in_blinding_key_sum = in_blinding_key_sum + _input.blinding_key if in_blinding_key_sum else _input.blinding_key
      in_blinding_key_sum += priv_data['priv_by_pub'][_input.address.pubkey.serialize()]
    output = IOput()
    output.fill(address, value, blinding_key = in_blinding_key_sum-out_blinding_key_sum,
      relay_fee=relay_fee, generator = default_generator_ser) #TODO relay fee should be distributed uniformly, privacy leak
    self.outputs.append(output)
    [output.generate() for output in self.outputs]
    self.sort_ioputs()
    self.verify()
    

  def calc_relay_fee(self, relay_fee_per_kb):
    inputs_num, outputs_num, excesses_num = len(self.inputs), len(self._destinations)+1, len(self.additional_excesses)
    input_size = 67+2
    output_size = 5366+2
    excess_size = 65+2
    estimated_size = 6+inputs_num*input_size + outputs_num*output_size + excesses_num*excess_size
    return int((estimated_size/1000.)*relay_fee_per_kb)


  def calc_new_outputs_fee(self, inputs_num=None, outputs_num=None):
    return ( (outputs_num if outputs_num else len(self.outputs)) -
             (inputs_num if inputs_num else len(self.inputs)) - 
             (1 if self.coinbase else 0))*output_creation_fee

  def sort_ioputs(self):
      self.inputs = sorted(self.inputs, key= lambda _input: _input.authorized_pedersen_commitment.serialize())
      self.outputs = sorted(self.outputs, key= lambda _output: _output.authorized_pedersen_commitment.serialize())

  #XXX: obsolete, to be deleted
  def calc_txid(self):
    if not len(self.inputs) or not len(self.outputs):
      raise Exception("Tx is not ready for id formation")
    self.sort_ioputs()
    to_hash = b''
    for _list in [self.inputs, self.outputs]:
        for ioput in _list:
            to_hash += ioput.pedersen_commitment.serialize()
    if not GLOBAL_TEST:
        raise NotImplemented 
    h = hashlib.new('sha256')
    h.update(to_hash)
    self.txid = h.digest()

  def serialize(self):
    if self.serialized:
      return self.serialized
    ret=b""
    ret += struct.pack("> H", len(self.inputs)) 
    for _input in self.inputs:
        s_i = _input.serialized_index
        ret += struct.pack("> H", len(s_i))
        ret += s_i

    ret += struct.pack("> H", len(self.outputs)) 
    for _output in self.outputs:
        s_o = _output.serialize()
        ret += struct.pack("> H", len(s_o))
        ret += _output.serialize()

    ret += struct.pack("> H", len(self.additional_excesses)) 
    for _excess in self.additional_excesses:
        s_e = _excess.serialize()
        ret += struct.pack("> H", len(s_e))
        ret += s_e
    if not GLOBAL_TEST['skip combined excesses']:
      raise NotImplemented
    self.serialized = ret
    return ret

  def deserialize(self, serialized_tx, skip_verification=False):
    self.serialized = None
    if len(serialized_tx)<2:
        raise Exception("Serialized transaction doesn't contain enough bytes for inputs array length")
    inputs_len_buffer, serialized_tx =serialized_tx[:2], serialized_tx[2:]
    (inputs_len,) = struct.unpack("> H", inputs_len_buffer) 
    for _input_index in range(inputs_len):
        if len(serialized_tx)<2:
          raise Exception("Serialized transaction doesn't contain enough bytes for input %s length"%_input_index)
        input_len_buffer, serialized_tx =serialized_tx[:2], serialized_tx[2:]
        (input_len,) = struct.unpack("> H", input_len_buffer)
        if len(serialized_tx)<input_len:
          raise Exception("Serialized transaction doesn't contain enough bytes for input %s"%_input_index)
        input_index_buffer, serialized_tx = serialized_tx[:input_len], serialized_tx[input_len:]

        if not GLOBAL_TEST['spend from mempool']:
            raise NotImplemented
        else:
          if not skip_verification:
            if (not input_index_buffer in self.txos_storage.confirmed):
              raise Exception("Unknown input index")
            self.inputs.append(self.txos_storage.confirmed[input_index_buffer])
          else:
            self.inputs.append(input_index_buffer)

    if len(serialized_tx)<2:
        raise Exception("Serialized transaction doesn't contain enough bytes for outputs array length")
    outputs_len_buffer, serialized_tx =serialized_tx[:2], serialized_tx[2:]
    (outputs_len,) = struct.unpack("> H", outputs_len_buffer)
    for _output_index in range(outputs_len):
      if len(serialized_tx)<2:
          raise Exception("Serialized transaction doesn't contain enough bytes for output %s length"%_output_index)
      output_len_buffer, serialized_tx =serialized_tx[:2], serialized_tx[2:]
      (output_len,) = struct.unpack("> H", output_len_buffer)
      if len(serialized_tx)<output_len:
          raise Exception("Serialized transaction doesn't contain enough bytes for output %s"%_output_index)
      output_buffer, serialized_tx = serialized_tx[:output_len], serialized_tx[output_len:]
      self.outputs.append(IOput(binary_object=output_buffer) )

    if len(serialized_tx)<2:
        raise Exception("Serialized transaction doesn't contain enough bytes for additional excesses array length")
    aes_len_buffer, serialized_tx =serialized_tx[:2], serialized_tx[2:]
    (aes_len,) = struct.unpack("> H", aes_len_buffer)
    for _ae in range(aes_len):
      if len(serialized_tx)<2:
          raise Exception("Serialized transaction doesn't contain enough bytes for additional excess %s length"%_ae)
      ae_len_buffer, serialized_tx =serialized_tx[:2], serialized_tx[2:]
      (ae_len,) = struct.unpack("> H", ae_len_buffer)
      if len(serialized_tx)<ae_len:
          raise Exception("Serialized transaction doesn't contain enough bytes for additional excess %s"%_ae)
      ae_buffer, serialized_tx = serialized_tx[:ae_len], serialized_tx[ae_len:]
      e = Excess()
      e.deserialize_raw(ae_buffer)
      self.additional_excesses.append(e )

    if not skip_verification:
      self.verify()
    if not GLOBAL_TEST['skip combined excesses']:
      raise NotImplemented

  def to_json(self):
    pass #TODO

  def from_json(self):
    pass #TODO

  def non_context_verify(self, block_height):
    #Actually we partially use context via block_height. Consider renaming.
    try:
      return verification_cache[(self.serialize, block_height)]
    except KeyError:
      pass

    assert is_sorted(self.inputs, key= lambda _input: _input.authorized_pedersen_commitment.serialize()), "Inputs are not sorted"
    assert is_sorted(self.outputs, key= lambda _output: _output.authorized_pedersen_commitment.serialize()), "Outputs are not sorted"

    for _input in self.inputs:
       assert _input.lock_height<block_height, "Timelocked input"

    #Check that there are no duplicated outputs
    #TODO probably authorized????
    assert len(set([_output.unauthorized_pedersen_commitment.serialize() for _output in self.outputs]))==len(self.outputs), "Duplicated output"

    coinbase_num=0
    
    output_apcs = []

    for output in self.outputs:
        assert output.verify(), "Nonvalid output"
        _o_index =output.serialized_index
        output_apcs.append(_o_index[:33])
        if output.version==0:
            coinbase_num+=1
            self.coinbase = output
    assert coinbase_num<2, "More then one coinbase"

    for excess in self.additional_excesses:
        assert excess.verify(), "Nonvalid excess"
        if not excess.message in output_apcs:
          return False
        else:
          output_apcs.remove(excess.message)

    left_side, right_side = [], []

    _t = PublicKey()

    # Transaction should contain either outputs (while it may not contain inputs)
    # or combined excesses (for transactions which only delete excesses)
    assert len(self.outputs) or len(self.combined_excesses), "Empty outputs"

    if len(self.inputs):
      _t.combine([_input.authorized_pedersen_commitment.to_public_key().public_key for _input in self.inputs])
      inputs_pedersen_commitment_sum = _t.to_pedersen_commitment()
      left_side.append(inputs_pedersen_commitment_sum)

    if len(self.outputs):
      _t.combine([_output.authorized_pedersen_commitment.to_public_key().public_key for _output in self.outputs])
      outputs_pedersen_commitment_sum = _t.to_pedersen_commitment()
      right_side.append(outputs_pedersen_commitment_sum)

      _t.combine([_output.address.pubkey.public_key for _output in self.outputs])
      outputs_excesses_sum = _t.to_pedersen_commitment()
      left_side.append(outputs_excesses_sum )

    if len(self.additional_excesses):
        _t.combine([excess.pubkey.public_key for excess in self.additional_excesses])
        additional_excesses_sum = _t.to_pedersen_commitment()
        left_side.append(additional_excesses_sum)

    if coinbase_num:
        minted_pc = PedersenCommitment(blinded_generator = default_generator)
        minted_pc.create(self.coinbase.value, b'\x00'*32)
        left_side.append(minted_pc)
     
     
    relay_fee = 0   
    for _output in self.outputs:
      if not _output.version==0:
        if _output.generator==default_generator_ser:
          relay_fee += _output.relay_fee

    new_outputs_fee = self.calc_new_outputs_fee()
    fee = relay_fee + new_outputs_fee

    negative_fee = False
    if fee<0:
      # It's okay, transaction has consumed so many inputs that it is profitable by itself
      # however we need to handle it manually: libsecp256k1 cannot handle negative value
      negative_fee = True
      fee = -fee

    if not fee==0:
      fee_pc = PedersenCommitment(blinded_generator = default_generator) #TODO think about fees for assets
      fee_pc.create(fee, b'\x00'*32)
      if negative_fee:
        left_side.append(fee_pc)
      else:
        right_side.append(fee_pc)

    checker = PedersenCommitment()
    # For transaction which contains coinbase only, both sides will be empty
    if len(left_side) or len(right_side):
      sum_to_zero = checker.verify_sum(left_side, right_side)
      assert sum_to_zero, "Non-zero Pedersen commitments summ"

    if not GLOBAL_TEST['skip combined excesses']:
      raise NotImplemented
    if self.coinbase:
      info = self.coinbase.info()
      assert info['exp'] == -1, "Non-transparent coinbase"
      # TODO Ugly ->`self.txos_storage.storage_space.blockchain.current_height`
      assert self.coinbase.lock_height >= block_height + coinbase_maturity,\
             "Wrong coinbase maturity timelock: %d should be %d"%(\
              self.coinbase.lock_height, block_height + coinbase_maturity)

    tx_skel = TransactionSkeleton(tx=self)
    assert len(tx_skel.serialize(rich_format=False))<50000, "Too big tx_skeleton"
    verification_cache[(self.serialize, block_height)] = True
    return True

    

  def verify(self, block_height = None, skip_non_context=False):
    """
     Transaction is valid if:
      0) inputs and outputs are sorted  (non context verification)
      1) all inputs are in utxo set
      2) no outputs are in utxo 
      3) all outputs are unique (exactly equal outputs are prohibited) (non context verification)
      3) all outputs are valid (non context verification)
      4) all additional excesses are valid (non context verification)
      5) sum(inputs)+sum(additional_excesses)+sum(outputs_excesses)+sum(minted_coins) == sum(outputs) + sum(fee) (non context verification)
      6) For each combined excess sum(digested_excesses) == combined_excess (non context verification)
      7) if coinbase is presented lock_height is set correctly (non context verification)
      8) size check : serialized tx_skeleton should have size less than 50k bytes (non context verification)
    """
    if not block_height:
      # Why +1 is here:
      # There are two general cases for verifying transactions:
      # 1) While generating new transaction. In this case transaction will be in the next block.
      # 2) While checking transaction in the block. In this case, current blockchain state is set 
      #     to prev block (prev to which we are checking) and block_height is current+1
      block_height = self.txos_storage.storage_space.blockchain.current_height + 1

    if not skip_non_context: # skip_non_context is used when non context verification for that tx was made earlier
      assert self.non_context_verify(block_height)

    if not GLOBAL_TEST['spend from mempool']:
        raise NotImplemented #XXX Check that all inputs in UTXO
    else:
        database_inputs = []
        assert len(set([_input.serialized_index for _input in self.inputs]))==len(self.inputs)
        for _input in self.inputs:
          index=_input.serialized_index
          if not index in self.txos_storage.confirmed:
              raise Exception("Spend unknown output")
          else:
              database_inputs.append(self.txos_storage.confirmed[index])
          self.inputs = database_inputs
           

    
    if not GLOBAL_TEST['spend from mempool']:
        raise NotImplemented #XXX Check that all inputs are in UTXOSet
    else:
        for _output in self.outputs:
          _o_index = _output.serialized_index
          if _o_index in self.txos_storage.confirmed:
              raise Exception("Create duplicate output")
          elif _o_index in self.txos_storage.mempool:
              pass #It's ok
          else:
              #previously unknown output, let's add to database
              self.txos_storage.mempool[_o_index] = _output


    if not GLOBAL_TEST['block_version checking']: 
        # Note, transactions where outputs have different block_versions are valid
        # by consensus rules. However, they cannot be included into any block.
        raise NotImplemented


    return True
    #[inputs_pedersen_commitment_sum, outputs_excesses_sum, additional_excesses_sum], [outputs_pedersen_commitment_sum, fee_pc])
    #sum(inputs)+sum(additional_excesses)+sum(outputs_excesses) == sum(outputs) + sum(fee)
    

  def merge(self, another_tx):
    self.serialized = None
    tx=Transaction(txos_storage = self.txos_storage, key_manager = self.key_manager) #TODO instead of key_manager, inputs info should be merged here
    tx.inputs=self.inputs+another_tx.inputs
    tx.outputs=self.outputs+another_tx.outputs
    tx.additional_excesses = self.additional_excesses + another_tx.additional_excesses
    if not GLOBAL_TEST['skip combined excesses']:
      raise NotImplemented
      #tx.combined_excesses = self.combined_excesses.update(another_tx.combined_excesses)
    tx.sort_ioputs()
    if not GLOBAL_TEST['spend from mempool']: 
      # If we merge transactions where the second spends outputs from the first, result is invalid
      # since we don't delete identical ioputs 
      raise NotImplemented
    assert tx.verify()
    return tx

  def __repr__(self):
    s=""
    s+="Inputs:\n"
    for _input in self.inputs:
       s+=" "*4+_input.__str__()+"\n"
    
    s+="Outputs:\n"
    for _output in self.outputs:
       s+=" "*4+_output.__str__()+"\n"
    return s

  @property
  def relay_fee(self):
    fee = 0
    for _output in self.outputs:
       fee += _output.relay_fee
    return fee

    



from leer.core.storage.merkle_storage import MMR
from secp256k1_zkp import PedersenCommitment, PublicKey
from leer.core.storage.default_paths import excesses_storage_path
import hashlib
import os
from leer.core.utils import sha256

class ExcessMMR(MMR):
      def sum(self, x1,x2):
        # each index is 33 bytes for commitments and 32 for hash
        pubkey1, hash1 = x1[:33], x1[33:65]
        pubkey2, hash2 = x2[:33], x2[33:65]
        pubkey1, pubkey2 = PublicKey(pubkey=pubkey1, raw=True), PublicKey(pubkey=pubkey2, raw=True)
        # consider pubkey1+pubkey2
        pk= PublicKey()
        pk.combine([pubkey1.public_key, pubkey2.public_key])
        first_part = pk.serialize()
        second_part = sha256(hash1+hash2)
        return first_part+second_part




class ExcessesStorage():
    '''
      Excesses tree:
        Each excess leaf obj is serialized excess, leaf index is excess_index.
    '''
    __shared_states = {}

    def __init__(self, storage_space, path):
      if not path in self.__shared_states:
        self.__shared_states[path]={}
      self.__dict__ = self.__shared_states[path]
      self.excesses = ExcessMMR("excesses", os.path.join(path, "confirmed"), clear_only=False)
      self.storage_space = storage_space
      storage_space.register_excesses_storage(self)
      
    def append_utxo(self, utxo):
      self.excesses.append(utxo.address.index, b"")


    def append_additional_excess(self, excess):
      self.excesses.append(excess.index, excess.serialize())

    def __contains__(self, serialized_index):
      return bool(self.excesses.get_by_hash(serialized_index))

    def remove(self,n):
      self.excesses.remove(n)

    def get_root(self):
      return self.excesses.get_root()

    def apply_tx_get_merkles_and_rollback(self, tx):
      for _o in tx.outputs:
          self.append_utxo(_o)
      for _e in tx.additional_excesses:
          self.append_additional_excess(_e)

      root = self.get_root()

      self.excesses.remove(len(tx.additional_excesses)+len(tx.outputs))
      return root

    #TODO bad naming. It should be apply block, or block_transaction
    def apply_tx(self, tx, new_state):
      for _o in tx.outputs:
          self.append_utxo(_o)
      for _e in tx.additional_excesses:
          self.append_additional_excess(_e)
      self.set_state(new_state)
      return len(tx.additional_excesses)+len(tx.outputs)

    def rollback(self, num_of_added_excesses, prev_state):
      self.excesses.remove(num_of_added_excesses)
      self.set_state(prev_state)

    def get_state(self):
      return self.excesses.get_state()

    def set_state(self, state):
      self.excesses.set_state(state)



from os.path import *
home = expanduser("~")
base_dir = join(home, ".leer")

def calc_paths(base_dir):
  blockchain = join(base_dir, "blockchain")
  txo_storage_path = join(blockchain, "txos_storage")
  excesses_storage_path = join(blockchain, "excesses_storage")
  headers_storage_path = join(blockchain, "headers_storage")
  blocks_storage_path = join(blockchain, "blocks_storage")

  wallet_path = join(base_dir, "wallet")
  key_manager_path = join(wallet_path, "default_wallet")
  utxo_index_path = join(base_dir, "meta", "utxo_index")

  return txo_storage_path, txo_storage_path, excesses_storage_path,\
         headers_storage_path, blocks_storage_path, wallet_path,\
         key_manager_path, utxo_index_path

txo_storage_path, txo_storage_path, excesses_storage_path,\
headers_storage_path, blocks_storage_path, wallet_path,\
key_manager_path, utxo_index_path = calc_paths(base_dir)



import os, lmdb


class UTXOIndex:
  '''
    Basically it is index [public_key -> set of unspent outputs with this public key] 
  '''
  __shared_states = {}

  def __init__(self, storage_space, path):
    if not path in self.__shared_states:
      self.__shared_states[path]={}
    self.__dict__ = self.__shared_states[path]
    self.directory = path
    if not os.path.exists(path): 
        os.makedirs(self.directory) #TODO catch
    self.env = lmdb.open(self.directory, max_dbs=10)
    with self.env.begin(write=True) as txn:
      self.main_db = self.env.open_db(b'main_db', txn=txn, dupsort=True, dupfixed=True) #TODO duplicate
    self.storage_space = storage_space
    self.storage_space.register_utxo_index(self)

  def _add(self, serialized_pubkey, utxo_hash_and_pc):
    with self.env.begin(write=True) as txn:
      txn.put( serialized_pubkey, utxo_hash_and_pc, db=self.main_db, dupdata=True)

  def add_utxo(self, output):
    self._add(output.address.pubkey.serialize(), output.serialized_index)

  def _remove(self, serialized_pubkey, utxo_hash_and_pc):
    with self.env.begin(write=True) as txn:
      txn.delete( serialized_pubkey, utxo_hash_and_pc, db=self.main_db)
    

  def remove_utxo(self, output):
    self._remove(output.address.pubkey.serialize(), output.serialized_index)


  def get_all_unspent(self, pubkey):
    return self.get_all_unspent_for_serialized_pubkey(pubkey.serialize())

  def get_all_unspent_for_serialized_pubkey(self, serialized_pubkey):
    with self.env.begin(write=False) as txn:
      cursor = txn.cursor(db=self.main_db)
      if not cursor.set_key(serialized_pubkey):
        return []
      else:
        return list(cursor.iternext_dup())
